# ==============================================================================
# sgxDbg.tcl --
#
#	SoftGuards eXtensions to Debug.
#
# Copyright (c) 1999 by T&E Soft, Stefan G. R. Finzel <finzel@osn.de>.
# All rights reserved.
#
# See the file "license.terms" for information on usage and redistribution of
# this file, and for a DISCLAIMER OF ALL WARRANTIES. 
#
# RCS: @(#) $Id: sgxDbg.tcl,v 1.7 2002/11/05 15:22:06 fis Exp $
# ==============================================================================

package provide sgxDbg 1.0

namespace eval ::sgxDbg:: {
    namespace export array_elements out stack var
}

# ------------------------------------------------------------------------------
proc ::sgxDbg::array_elements {vars {idx {}}} {
    # --------------------------------------------------------------------------
    # Display list of _global_ variables or list of elements of list of _global_ arrays
    # --------------------------------------------------------------------------
    foreach var ${vars} {
        # FRINK: nocheck
        global [set var]

        puts stderr {}
        set idxs ${idx}
        if {![llength ${idx}]} {
            # FRINK: nocheck
            set idxs [lsort -dictionary [array names [set var]]]
        } elseif {[string first {*} ${idx}] || [string first {?} ${idx}]} {
            # FRINK: nocheck
            set idxs [lsort -dictionary [array names [set var] ${idx}]]
        } else {
            puts stderr "\t${var}\t[subst $${var}]"
            return
        }
        foreach field ${idxs} {
            catch {puts stderr "\t${var}(${field})\t'[subst $${var}(${field})]'"}
        }
    }
}

# ------------------------------------------------------------------------------
proc ::sgxDbg::out {args} {
    # --------------------------------------------------------------------------
    puts stderr [join ${args} { / }]
}

# ------------------------------------------------------------------------------
proc ::sgxDbg::stack {{level -1} {par 1} {calls 0}} {
    # --------------------------------------------------------------------------
    # determine level proc, parameter value or definitions and calltree
    # --------------------------------------------------------------------------
    set proclevel [info level]

    if {${level} < 0} {
        set level [expr {${proclevel} + ${level}}]
    }

    set tab {}
    set stack {}
    # FRINK: set runlevel
    for {set runlevel 1} {(${runlevel} <= ${level}) &&(${level} <= ${proclevel})} {incr runlevel} {
        if {${par}} {
            set call [info level ${runlevel}]
        } else {
            set call "[lindex [info level ${runlevel}] 0]{[info args [lindex [info level\
              ${runlevel}] 0]]}"
        }
        if {${calls}} {
            append stack "\n${tab}proc ${call}"
            append tab "  "
        } else {
            set stack "proc ${call}"
        }
    }
    puts stderr ${stack}
}

# ------------------------------------------------------------------------------
proc ::sgxDbg::var {var {idx {}} {type w} {call ::sgxDbg::var_puts}} {
    # --------------------------------------------------------------------------
    # Trace variables or arrays
    # --------------------------------------------------------------------------
    # FRINK: nocheck
    global [set var]

    if {![llength ${idx}]} {
        trace variable [set var] ${type} ${call}
    } else {
        trace variable [set var(${idx})] ${type} ${call}
    }
    # FRINK: ${var}
}

# ------------------------------------------------------------------------------
proc ::sgxDbg::var_puts {var idx type} {
    # --------------------------------------------------------------------------
    # Output routine tracing variables or arrays
    # --------------------------------------------------------------------------
    # FRINK: ${idx} unused
    upvar ${var} ref

    switch -exact -- ${type} {
        r {
                puts stderr "read\t${ref}\t[subst $${ref}]"
            }
        u {
                puts stderr "unset\t${ref}"
            }
        w {
                puts stderr "write\t${ref}\t[subst $${ref}]"
            }
        default {
                puts stderr "error\tinvalid switch type '${type}'"
            }
    }
}

# ==============================================================================
