# ==============================================================================
# sgxTVC.tcl --
#
#	sgxTVC: SoftGuards eXtension for Testing Variable Consistency
#
# Copyright (c) 1999 by T&E Soft, Stefan G. R. Finzel <finzel@osn.de>.
# All rights reserved.
#
# See the file "license.terms" for information on usage and redistribution of
# this file, and for a DISCLAIMER OF ALL WARRANTIES. 
#
# RCS: @(#) $Id: sgxTVC.tcl,v 1.26 2002/11/08 18:31:08 fis Exp $
# ==============================================================================
# code has to be validated and beautified to get best results
#        use e.g.: TclPro, frink
# ==============================================================================

rename proc _proc
_proc proc {name args body} {
    global sgxTVC
    # warn only true redefinitions
    if {[llength [info commands ${name}]] && ([string compare [info args ${name}] ${args}] ||\
  [string compare [info body ${name}] ${body}])} {
	# puts stderr "proc\twarning\t[info script] redefines proc ${name}"	
    }
    set tmpname {}
    if {[regexp -- {::([^:]+)$} ${name} UNUSED tmpname] && (0 <= [lsearch -exact [info commands]\
  ${tmpname}])} {
	puts stderr "proc\twarning\t[info script] proc ${name} hiding command ${tmpname}"	
    }
    _proc ${name} ${args} ${body}
}

# sgxTVC configuration
#   tests        tests to perform
array set sgxTVC {
    tests        {_refs_ _defs_ _uses_}
    global       {argv0}
}

# support old tcl versions too
if {8.0 <= ${::tcl_version}} {
    package provide sgxTVC 1.0

    namespace eval ::sgxTVC:: {
        namespace export _Files_ _Pkgs_ _defs_ _refs_ _uses_
    }
}

# ==============================================================================

# ------------------------------------------------------------------------------
# ::sgxTVC::_Files_ --
#                      test all procedures within a list of files
# Arguments:
#        filenames:    list of filenames to source and test;
# Results:
#        integer       total counter of failures
# ------------------------------------------------------------------------------
proc ::sgxTVC::_Files_ {{filenames {}}} {
    # --------------------------------------------------------------------------
    # TODO: use interp
    # --------------------------------------------------------------------------
    global sgxTVC

    # initialize globals required by sourced files
    foreach var $sgxTVC(global) {
        global [set var]
    }

    set vars [info vars]
    set cmds [info commands]
    foreach filename [lsort ${filenames}] {
        puts stdout ${filename}
        if {[catch {source ${filename}} err]} {
            puts stderr ${err}
        }
    }
    set scan_cmds {}
    foreach cmd [info commands] {
        if {[lsearch -exact ${cmds} ${cmd}] < 0} {
            lappend scan_cmds ${cmd}
        }
    }
    set cmds [lsort ${scan_cmds}]
    # check procs now
    set fail 0
    foreach cmd $sgxTVC(tests) {
        incr fail [::sgxTVC::${cmd} ${cmds} ${vars}]
    }
    return ${fail}
}

# ------------------------------------------------------------------------------
# ::sgxTVC::_Pkgs_ --
#                     test all matching packages
# Arguments:
#        pkgs_re:     regular expression to select all packages to test
# Results:
#        integer      total counter of failures
# ------------------------------------------------------------------------------
proc ::sgxTVC::_Pkgs_ {{pkgs_re {.*}}} {
    # --------------------------------------------------------------------------
    # TODO: use interp
    # --------------------------------------------------------------------------
    global prg tcl_version sgxTVC

    update
    set pkgs {}
    # FRINK: set prg(app,name); to get and add the main programs package name every time
    catch {lappend pkgs $prg(app,name)}
    # do not: package provide a_bogus_package_name or proc ::a_bogus_package_name::
    catch {package require a_bogus_package_name}
    foreach pkg [package names] {
        if {[regexp -- ${pkgs_re} ${pkg} UNUSED]} {
            lappend pkgs ${pkg}
        }
    }

    # handle old tcl/tk versions too
    set delimiter {}
    if {8.0 <= ${tcl_version}} {
        set delimiter {::}
    }

    set fail 0
    foreach pkg [lsort ${pkgs}] {
        puts stdout ${pkg}
        catch {package require ${pkg} 1.0}
        # many of my namespaces use an initialisation proc for autoloading
        #   and dynamically creating procs at run time 
        catch {${delimiter}${pkg}${delimiter}Init}
        # get namespace commands and variables
        set cmds [lsort [info commands ${delimiter}${pkg}${delimiter}*]]
        set vars [info vars ${delimiter}*]
        # check namespace procs now
        foreach cmd $sgxTVC(tests) {
            incr fail [::sgxTVC::${cmd} ${cmds} ${vars}]
        }
        # remove package
        catch {package forget ${pkg}}
    }
    return ${fail}
}

# ------------------------------------------------------------------------------
# ::sgxTVC::_defs_ --
#                     test for existing variable definitions
# Arguments:
#        procs:       list of procs to test
#        vars:        list of known variables names         
# Results:
#        integer      total counter of failures
# ------------------------------------------------------------------------------
proc ::sgxTVC::_defs_ {procs vars} {
    # --------------------------------------------------------------------------
    set def_cmds {{array anymore} {array donesearch} {array exists} {array get} {array names}\
      {array nextelement} {array size} {array startsearch} {incr} {info exists} {unset}}
    # make the next line a comment to skip its testing 
    lappend def_cmds {append} {lappend}

    set fail 0
    set info_cmds [info commands]
    foreach cmd ${procs} {
        if {[catch {
            set proc_vars [info args ${cmd}]
            set proc_body [info body ${cmd}]
        }]} {
            continue
        }
        # got args; add them to given variables
        set proc_vars [concat ${proc_vars} ${vars}]

        set proc_body_rest ${proc_body}
        while {[regexp -- "\\W(([join ${def_cmds} {|}]) \[:A-Za-z\]+\[:0-9A-Za-z_\]+)(.)"\
          ${proc_body_rest} UNUSED proc_cmd_var UNUSED next_char]} {
            set proc_var_pos [string first ${proc_cmd_var} ${proc_body_rest}]
            set proc_var [lindex ${proc_cmd_var} end]
            if {0 <= [lsearch -exact ${proc_vars} ${proc_var}]} {
                # already defined
            } elseif {![string compare ${next_char} {(}]} {
                # skip arrays
            } elseif {![regexp -- "( ${proc_var}\\W|\{${proc_var} )" [string range ${proc_body} 0\
              [string first "${proc_cmd_var}${next_char}" ${proc_body}]] UNUSED]} {
                # undefined
                puts stderr "${cmd}\terror\t${proc_var}\tvariable undefined used by\
                  [lindex ${proc_cmd_var} 0]"
                incr fail
                # add to defined variables 
                lappend proc_vars ${proc_var}
            } else {
                # add to defined variables
                lappend proc_vars ${proc_var}
            }
            # FRINK: ${proc_vars} will be used within the loop
            if {0 <= [lsearch -exact ${info_cmds} ${proc_var}]} {
                puts stderr "${cmd}\twarning\t${proc_var}\tbad style variable"
            }
            # remove scanned
            set proc_body_rest [string range ${proc_body_rest} [expr {${proc_var_pos} +\
              [llength ${proc_cmd_var}]}] end]
        }
    }
    return ${fail}
}

# ------------------------------------------------------------------------------
# ::sgxTVC::_refs_ --
#                     test if a variable is referenced
# Arguments:
#        procs:       list of proc to test
#        vars:        list of known variables names         
# Results:
#        integer      total counter of failures
# ------------------------------------------------------------------------------
proc ::sgxTVC::_refs_ {procs vars} {
    # --------------------------------------------------------------------------
    set fail 0
    set info_cmds [info commands]
    foreach cmd ${procs} {
        if {[catch {
            set proc_vars [info args ${cmd}]
            set proc_body [info body ${cmd}]
        }]} {
            continue
        }
        # got args; add them to given variables
        set proc_vars [concat ${proc_vars} ${vars}]

        while {[regexp -- {\$\{?[:A-Za-z]+[:0-9A-Za-z_]+\}?} ${proc_body} proc_var_ref]} {
            set proc_var_pos [string first ${proc_var_ref} ${proc_body}]
            set proc_var_last [string last "\n" ${proc_body} ${proc_var_pos}]
            regexp -- {[:0-9A-Za-z_]+} ${proc_var_ref} proc_var
            if {0 <= [lsearch -exact ${proc_vars} ${proc_var}]} {
                # already defined
            } elseif {![regexp -- "( ${proc_var}\\W|\{${proc_var} )" [string range ${proc_body} 0\
              ${proc_var_last}] UNUSED]} {
                # undefined
                puts stderr "${cmd}\terror\t${proc_var}\tvariable undefined"
                incr fail
                lappend proc_vars ${proc_var}
            } else {
                # add to referenced variables
                lappend proc_vars ${proc_var}
                # FRINK: ${proc_vars} used in loop
            }
            if {0 <= [lsearch -exact ${info_cmds} ${proc_var}]} {
                puts stderr "${cmd}\twarning\t${proc_var}\tbad style variable"
            }
            # remove $ by ~
            set proc_body "[string range ${proc_body} 0 [expr {${proc_var_pos} - 1}]]~[string\
              range ${proc_body} [expr {${proc_var_pos} + 1}] end]"
        }
    }
    return ${fail}
}

# ------------------------------------------------------------------------------
# ::sgxTVC::_uses_ --
#                     test usage of variables
# Arguments:
#        procs:       list of proc to test
#        vars:        list of known variables names         
# Results:
#        integer      total counter of failures
# ------------------------------------------------------------------------------
proc ::sgxTVC::_uses_ {procs vars} {
    # --------------------------------------------------------------------------
    # TODO: better support for commands: regexp/regsub/scan ...
    # --------------------------------------------------------------------------
    set def_cmds [list {global} {upvar} {variable}]

    set fail 0
    set info_cmds [info commands]
    foreach cmd ${procs} {
        # 1. get external variables
        if {[catch {
            set proc_vars [info vars ::*]
            set proc_body [info body ${cmd}]
        }]} {
            continue
        }
        # and add to given variables
        set proc_vars [concat ${proc_vars} ${vars}]
        set initial_proc_vars ${proc_vars}

        # 2. create list of local variables
        while {[regexp -- "\\s(([join ${def_cmds} {|}]) +\[A-Za-z\]+\[0-9A-Za-z_\]+)" ${proc_body}\
          UNUSED proc_cmd_var]} {
            set proc_var_pos [string first ${proc_cmd_var} ${proc_body}]
            lappend proc_vars [lindex ${proc_cmd_var} end]
            set proc_body [string range ${proc_body} [expr {${proc_var_pos} +\
              [llength ${proc_cmd_var}]}] end]
        }

        set proc_body [info body ${cmd}]

        # 3. check usage of proc arguments
        foreach proc_arg [info args ${cmd}] {
            if {![regexp -- "\\$\\{?${proc_arg}\\}?\\(?" ${proc_body}]} {
                # unused
                puts stderr "${cmd}\terror\t${proc_arg}\targument unused"
                incr fail
            }
            if {0 <= [lsearch -exact ${info_cmds} ${proc_arg}]} {
                puts stderr "${cmd}\twarning\t${proc_arg}\tbad style argument"
            }
        }

        # 4. check usage of direct defined variables
        set cmds {{array set} {append} {incr} {lappend} {\s *set} {lappend}}
        set ref_cmds [list {array names} {array exists} {info exists}]
        while {[regexp -- "\\s(([join ${cmds} {|}])( #\[0-9\]+)? +\[:A-Za-z\]+\[:0-9A-Za-z_\]+)"\
          ${proc_body} UNUSED proc_cmd_var]} {
            set proc_var_pos [string first ${proc_cmd_var} ${proc_body}]
            set proc_var [lindex ${proc_cmd_var} end]

            if {0 <= [lsearch -exact ${proc_vars} ${proc_var}]} {
                # already used
            } elseif {[regexp -- "\\$\\{?${proc_var}\\}?\\(?" ${proc_body} UNUSED]} {
                # add to used variables
                lappend proc_vars ${proc_var}
            } elseif {[regexp -- "\[^:0-9A-Za-z_\]([join ${ref_cmds}\
              {|}]) +${proc_var}\[^:0-9A-Za-z_\]" ${proc_body}]} {
                # add to used variables
                lappend proc_vars ${proc_var}
            } else {
                # unused
                puts stderr "${cmd}\terror\t${proc_var}\tvariable unused"
                incr fail
                lappend proc_vars ${proc_var}
            }
            if {0 <= [lsearch -exact ${info_cmds} ${proc_var}]} {
                puts stderr "${cmd}\twarning\t${proc_var}\tbad style variable"
            }
            # remove variable
            set proc_body [string range ${proc_body} [expr {${proc_var_pos} +\
              [llength ${proc_cmd_var}]}] end]
        }

        # 5. check usage of indirect defined variables
        ### at least this part has to be rewritten to avoid my dirty tricks ###
        set proc_vars ${initial_proc_vars}
        set proc_body [info body ${cmd}]
        lappend proc_vars UNUSED
        set cmds {{catch} {global} {regexp} {regsub} {scan} {trace variable} {trace vdelete}\
          {trace vinfo} {variable}}
        # skip commands 
        set skip_cmds {{trace variable} {trace vdelete} {trace vinfo}}

        # referencing a var is allowed too
        set ref_cmds {{append} {array [a-z]+} {incr} {info exists} {lappend} {set} {unset}}

        while {[regexp -- "\\s(([join ${cmds} {|}]) \[^;\\n\\r\]+)" ${proc_body} UNUSED\
          proc_cmd_var]} {
            set proc_var_pos [string first ${proc_cmd_var} ${proc_body}]
            if {![regexp -- "^([join ${skip_cmds} {|}]) " ${proc_cmd_var} UNUSED]} {

                # don't get confused by closing brackets from command substitution
                regsub -all -- {\[[^\]]+\]} ${proc_cmd_var} {@@@} proc_cmd_var
                regsub -all -- {\$\{[^\}]+\}} ${proc_cmd_var} {@@@} proc_cmd_var
                regsub -all -- {\{[^\}]+\}} ${proc_cmd_var} {@@@} proc_cmd_var
                regsub -all -- {"[^\"]"} ${proc_cmd_var} {@@@} proc_cmd_var
                regsub -all -- {\[|\]|\{|\}} ${proc_cmd_var} {} proc_cmd_var

                # skip all arguments not containing indirect defined var names
                switch -exact -- [lindex ${proc_cmd_var} 0] {
                    regexp {
                            # FRINK: set offset by for
                            for {set offset 1} {[string match {-*} [lindex ${proc_cmd_var}\
                              ${offset}]]} {incr offset} {
                                if {![string compare {--} [lindex ${proc_cmd_var} ${offset}]]} {
                                    incr offset
                                    break
                                }
                            }
                            incr offset 2
                        }
                    regsub {
                            for {set offset 1} {[string match {-*} [lindex ${proc_cmd_var}\
                              ${offset}]]} {incr offset} {
                                if {![string compare {--} [lindex ${proc_cmd_var} ${offset}]]} {
                                    incr offset
                                    break
                                }
                            }
                            incr offset 3
                        }
                    scan {
                            set offset 3
                        }
                    catch {
                            set offset 2
                        }
                    upvar {
                            set offset end
                        }
                    default {
                            set offset 1
                        }
                }

                foreach proc_var [lrange ${proc_cmd_var} ${offset} end] {
                    if {![regexp -nocase -- {^([_:0-9A-Z]+)} ${proc_var} UNUSED proc_var]} {
                        continue
                    }

                    if {0 <= [lsearch -exact ${proc_vars} ${proc_var}]} {
                        # already used
                    } elseif {![regexp -- "(\\$\\{?${proc_var}\\}?\\(?|([join ${ref_cmds}\
                      {|}]) ${proc_var}\\(?)" ${proc_body} UNUSED]} {
                        # unused
                        puts stderr "${cmd}\terror\t${proc_var}\tvariable unused"
                        incr fail
                        lappend proc_vars ${proc_var}
                    } else {
                        # add to used variables
                        lappend proc_vars ${proc_var}
                    }
                    if {0 <= [lsearch -exact ${info_cmds} ${proc_var}]} {
                        puts stderr "${cmd}\twarning\t${proc_var}\tbad style variable"
                    }
                }
            }
            # remove variable
            set proc_body [string range ${proc_body} [expr {${proc_var_pos} +\
              [llength ${proc_cmd_var}]}] end]
        }
    }
    return ${fail}
}

# ==============================================================================

# examples of usage (self test):

# ::sgxTVC::_Files_ [file join .. test test_tvc.dat]
# ::sgxTVC::_Pkgs_ ::sgxTVC
# ::sgxTVC::_rsm_ {::sgxTVC::_Files_ ::sgxTVC::_Pkgs_ ::sgxTVC::_defs_ ::sgxTVC::_refs_\
  ::sgxTVC::_uses_}

# ==============================================================================

